from typing import Optional, List
from pydantic import BaseModel, Field, HttpUrl, validator
from enum import Enum




class EmbeddingModelName(str, Enum):
    ada002 :str = "text-embedding-ada-002"
    
    def __str__(self):
        return str(self.value)    

    
class EmbeddingConfig(BaseModel):
    model : EmbeddingModelName =  Field(EmbeddingModelName.ada002, description="The name of the model to use for embedding the collection content.")
    


from typing import List
from .org import Org
from .collection import Collection
from .models.user import User, ApiKey
from .jiggybase_session import JiggyBaseSession

    

class JiggyBase():

    def __init__(self):
        self.session = JiggyBaseSession()     
        
    def orgs(self) -> List[Org]:
        """
        return all Orgs that the user is a member of
        """
        resp = self.session.get('/orgs')
        return [Org(self.session, **i) for i in resp.json()]

    def get_org(self, name_or_id) -> Org:
        """
        get org by name or id
        raises Exception if an exact match for name is not found
        """
        orgs = [t for t in self.session.get('/orgs').json() if t['name'] == name_or_id or t['id'] == name_or_id]
        if len(orgs):
            return Org(self.session, **orgs[0])
        raise Exception(f'Org "{name_or_id}" not found')


    def api_keys(self) -> List[ApiKey]:
        """
        return user's api keys
        """
        return [ApiKey(**i) for i in self.session.get('/apikey').json()]


    def authenticated_user(self) -> User:
        """
        return the authenticated user's User object
        """
        return User(**self.session.get("/users/current").json())


    def create_org(self, name : str) -> Org:
        """
        create an Org
        """
        resp = self.session.post("/orgs", json={"name":name})
        return Org(self.session, **resp.json())

    def collections(self) -> List[Collection]:
        """
        return all Collections in all Orgs that the user is a member of
        """
        collections = []
        for org in self.orgs():
            for collection in org.collections():
                collections.append(collection)
        return collections

    def collection_names(self) -> List[Collection]:
        """
        return the unique collection name for all collections in all Orgs that the user is a member of
        """
        names = []
        for org in self.orgs():
            for collection in org.collections():
                names.append(collection.name)
        return names
            
    def collection(self, name : str) -> Collection:
        for org in self.orgs():
            for collection in org.collections():
                if collection.name == name:
                    return collection
        raise ValueError(f'Collection "{name}" not found')



import jiggybase


from jiggybase.models import (
    ApiKey,
    ChatModelName,
    ChunkConfig,
    Collection,
    CollectionPatchRequest,
    CollectionPostRequest,
    EmbeddingConfig,
    EmbeddingModelName,
    ExtractMetadataConfig,
    Org,
    OrgMember,
    OrgMemberPostRequest,
    OrgPatchRequest,
    OrgPostRequest,
    OrgRole,
    PatchPluginOAuthConfigRequest,
    PluginAuthConfigOAuth,
    PluginAuthType,
    PluginBearerTokenConfig,
    PluginConfig,
    User,
    UserPostPatchRequest,
    UserPostRequest,
    PromptTask
)




if __name__ == "__main__":
    jb = jiggybase.JiggyBase()
    orgs = jb.orgs()
    for org in orgs:
        print(f'Org: {org}')
        print(org.prompt_tasks())
        for c in org.collections():
            print(f'   Collection: {c}')    
    collections = jb.collections()


import os
from typing import Optional, List, Iterator
from pydantic import BaseModel, Field, BaseConfig, HttpUrl
from enum import Enum
from .models import collection, CollectionChatConfig, PatchCollectionChatConfig
from .jiggybase_session import JiggyBaseSession
from .models import UpsertResponse,  Query, QueryRequest, QueryResponse, UpsertRequest, Document, DocumentChunk, DeleteRequest, DeleteResponse, DocumentMetadataFilter, DocChunksResponse
from .models import Message, CompletionRequest, ChatCompletion, ChatUsage
from .chat_stream import extract_content_from_sse_bytes

from typing import Union, List
class PluginAuthType(Enum):
    bearer :str = "bearer"
    none   :str = "none"
    oauth  :str = "oauth"

        
class CollectionPostRequest(BaseModel):
    """
    Used to create a new Collection service
    """
    name:         str                      = Field(description="The globally unique hostname for the collection. Subject to DNS naming rules. ")
    display_name: Optional[str]            = Field(description="The human friendly display name for the collection.")
    description:  Optional[str]            = Field(description="A description of the collection.")
    plugin_auth:  Optional[PluginAuthType] = Field(default=PluginAuthType.bearer, description="The authentication type to use for this collection's ChatGPT plugin.")


class CollectionPatchRequest(BaseModel):
    """
    Used to modify an existing service
    """
    description:  Optional[str] = Field(description="A description of the collection.")
    plugin_auth:  Optional[str] = Field(description='The auth plugin to use for this org.')
    display_name: Optional[str] = Field(description="The human friendly display name for the collection.")



class PluginAuthConfigOAuth(BaseModel):
    """
    The Plugin Oauth configuration as managed by GPT-Gateway
    """    
    client_url:                 HttpUrl     = Field(description="ChatGPT will direct the user’s browser to this url to log in to the plugin")
    authorization_url:          HttpUrl     = Field(description="After successful login ChatGPT will complete the OAuth flow by making a POST request to this URL")
    scope:                      str         = Field(description="The scope used for the OAuth flow")
    client_id:                  str         = Field(unique=True, index=True, description="The client id to send to OpenAI for the plugin")
    client_secret:              str         = Field(description="The client secret to send to OpenAI for the plugin")
    openai_verification_token:  str         = Field(description="The verification token specified by OpenAI to configure in the plugin")

class PatchPluginOAuthConfigRequest(BaseModel):
    openai_verification_token: str  

        
class Collection(collection.Collection):
    """
    derived from models.collection.Collection data model for purposes of adding management methods
    """
    class Config(BaseConfig):
        extra = "allow"

    def __init__(self, session, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.session = session
        self.plugin_session = JiggyBaseSession(host=f'https://{kwargs["fqdn"]}', api='')
        self.chat_session = JiggyBaseSession(host=session.host, api='v1')
        
    def set_description(self, description:str) -> "Collection":
        """
        Update an existing collection using its ID and the provided description.
        """
        patch_request = CollectionPatchRequest(description=description)
        rsp = self.session.patch(f"/orgs/{self.org_id}/collections/{self.id}", model=patch_request)
        return Collection(**rsp.json())

    def set_oauth_verification_token(self, openai_verification_token: str) -> PluginAuthConfigOAuth:
        """
        Set the OpenAI verification token for this collection's plugin.
        This is the token that OpenAI provides during while registering a plugin configured to use Oauth.
        """
        patch_request = PatchPluginOAuthConfigRequest(openai_verification_token=openai_verification_token)
        rsp = self.session.patch(f"/orgs/{self.org_id}/collections/{self.id}/plugin_oauth", model=patch_request)
        return PluginAuthConfigOAuth(**rsp.json())

    def plugin_oauth_config(self) -> PluginAuthConfigOAuth:
        """
        Get the OAuth configuration for this collection's plugin.
        The client_id and client_secret returned here are configured with OpenAI while 
        registering a plugin configured to use Oauth.
        """
        rsp = self.session.get(f"/orgs/{self.org_id}/collections/{self.id}/plugin_oauth")
        return PluginAuthConfigOAuth(**rsp.json())

    def delete(self):
        """
        Delete a collection.  Warning: this is permanent.
        """
        self.session.delete(f"/orgs/{self.org_id}/collections/{self.id}")


    def get_chat_config(self) -> CollectionChatConfig:
        """
        Get the chat configuration for this collection.
        """
        rsp = self.session.get(f"/orgs/{self.org_id}/collections/{self.id}/chat_config")
        return collection.CollectionChatConfig(**rsp.json())

    def update_chat_config(self, model: str, prompt_task_id: int) -> CollectionChatConfig:
        """
        Update the chat configuration for this collection.
        """
        rsp = self.session.patch(f"/orgs/{self.org_id}/collections/{self.id}/chat_config/{model}", model=PatchCollectionChatConfig(prompt_task_id=prompt_task_id))
        return CollectionChatConfig(**rsp.json())

 
    def upsert_file(self, file_path: str, mimetype: str = None) -> UpsertResponse:
        """
        Add a file to the collection.
        """
        if not os.path.exists(file_path):
            raise ValueError("File not found")
             
        with open(file_path, "rb") as file:
            files = {"file": (os.path.basename(file_path), file, mimetype)}
            rsp = self.plugin_session.post("/upsert-file", files=files)
        return UpsertResponse.parse_obj(rsp.json())

    def upsert(self, documents: List[Document]) -> UpsertResponse:
        """
        Add a list of Document objects to the collection.
        """
        upsert_request = UpsertRequest(documents=documents)
        rsp = self.plugin_session.post("/upsert", model=upsert_request)
        return  UpsertResponse.parse_obj(rsp.json())
    
    def query(self, queries: Union[str, List[str], Query], top_k : int = 10) -> QueryResponse:
        """
        Query the collection returning the top_k results for each query.
        queries can be either a single string, a list of strings, or a list of Query objects.
        if it is a string or list of strings, the specified top_k will be used for each of the queries.
        Returns a QueryResponse object.
        """
        if isinstance(queries, str):
            queries = [Query(query=queries, top_k=top_k)]
        elif isinstance(queries, Query):
            queries = [Query(query=q, top_k=top_k) for q in queries]
        qr = QueryRequest(queries=queries)
        rsp = self.plugin_session.post("/query", model=qr)
        return  QueryResponse.parse_obj(rsp.json())

    def get_doc(self, id: str) -> list[DocumentChunk]:
        """
        Get a document by id
        """
        rsp = self.plugin_session.get(f"/docs/{id}")
        return [DocumentChunk.parse_obj(c) for c in rsp.json()]
    
    def get_chunks(self, 
                   start: int = 0, 
                   limit: int = 10, 
                   reverse: bool = True) -> List[DocumentChunk]:
        """
        low level interface for iterating through the chunks in a collection
        start - Offset of the first result to return
        limit - Number of results to return starting from the offset
        reverse - Reverse the order of the items returned
        """
        params = {"start": start, "limit": limit, "reverse": reverse}
        rsp = self.plugin_session.get("/chunks", params=params)
        return [DocumentChunk.parse_obj(chunk) for chunk in rsp.json()]


    def delete_docs(self, 
                    ids                      : Optional[List[str]] = None, 
                    document_metadata_filter : Optional[DocumentMetadataFilter] = None, 
                    delete_all               : Optional[bool] = False) -> DeleteResponse:
        """
        Delete items in the collection by document id's or document metadata filter.
        A delete_all option is also provided to delete all documents in the collection.
        """
        delete_request = DeleteRequest(ids=ids, filter=document_metadata_filter, delete_all=delete_all)
        rsp = self.plugin_session.delete("/delete", model=delete_request)
        return DeleteResponse.parse_obj(rsp.json())


    def _chat_completion(self, 
                         messages: List[Message],
                         model = 'gpt-3.5-turbo',
                         max_tokens = None,
                         temperature = .1) -> ChatCompletion:
        """
        low level interface for chat completion
        The input here mirrors openai chat completion parameters, while the output is different
        """
        cr = CompletionRequest(model       = f'{self.name}_{model}', 
                               messages    = messages,
                               max_tokens  = max_tokens,
                               temperature = temperature,
                               stream      = False)
        rsp = self.chat_session.post("/chat/completions", model=cr)          
        try:            
            return ChatCompletion.parse_obj(rsp.json())
        except:
            return ChatUsage.parse_obj(rsp.json())  # transitional hack

    def _chat_completion_stream_str(self, 
                         messages: List[Message],
                         model = 'gpt-3.5-turbo',
                         max_tokens = None,
                         temperature = .1) -> Iterator[str]:
        """
        low level interface for chat completion with streaming
        yields the model output as an iteration of strings
        """
        cr = CompletionRequest(model       = f'{self.name}_{model}', 
                               messages    = messages,
                               max_tokens  = max_tokens,
                               temperature = temperature,
                               stream      = True)
        rsp = self.chat_session.post("/chat/completions", model=cr, stream=True)
        for line in rsp.iter_lines():
            if line:  # filter out keep-alive newlines        
                yield extract_content_from_sse_bytes(line)
                


#! /usr/bin/env python3

import os
import sys
import argparse
import jiggybase

jb = jiggybase.JiggyBase()

JIGGYBASE_ORG = os.environ.get("JIGGYBASE_ORG")
JIGGYBASE_COLLECTION = os.environ.get("JIGGYBASE_COLLECTION")


def upload_file(collection : jiggybase.models.Collection, filename : str):
    print(f'Uploading {filename}')
    try:
        upsert_rsp = collection.upsert_file(filename)
    except Exception as e:            
        print(f'error on {filename}: {e}')
        return
    doc_id = upsert_rsp.ids[0]
    dcl =  collection.get_doc(doc_id)
    text_len = len(" ".join([dc.text for dc in dcl]))
    title = dcl[0].metadata.title if dcl[0].metadata.title else "Unnown Title"
    print(f'Processed {filename}: "{title}"  {text_len//1024} KB text ({len(dcl)} chunks)')
    

def upload_directory(collection : jiggybase.models.Collection, dirname : str):
    for fn in os.listdir(dirname):
        fn = os.path.join(dirname, fn)
        upload_file(collection, fn)


epilog = "If neither '--file' nor '--dir' options are provided, the script will automatically process other arguments as a file or directory"


def main():
    parser = argparse.ArgumentParser(description="Upload a file or directory to a JiggyBase collection",  epilog=epilog )    
    parser.add_argument("--org", type=str, help="The name of your JiggyBase organization.  Alternatively, set JIGGYBASE_ORG environment variable or be a member of a single Org.")
    parser.add_argument("--collection", type=str, help="The name of your JiggyBase collection. Alternatively, set JIGGYBASE_COLLECTION environment variable or have a single collection in your org.")
    parser.add_argument("--dir", type=str, help="The directory you want to upload.")
    parser.add_argument("--file", type=str, help="The file you want to upload")

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    parsed_args, unknown_args = parser.parse_known_args(sys.argv[1:])
    if not parsed_args.org:
        orgs = jb.orgs()
        if len(orgs) > 1:
            print("Please provide an organization name using the --org option")
            sys.exit(1)
        elif JIGGYBASE_ORG:
            org = jb.get_org(JIGGYBASE_ORG)
        else:
            org = orgs[0]
    else:
        org = jb.get_org(parsed_args.org)

    if not parsed_args.collection:
        collections = org.collections()
        if len(collections) > 1:
            print("Please provide a collection name using the --collection option")
            sys.exit(1)
        elif JIGGYBASE_COLLECTION:
            collection = org.collection(JIGGYBASE_COLLECTION)
        else:
            collection = collections[0]
    else:
        collection = org.collection(parsed_args.collection)

    if parsed_args.dir:
        upload_directory(collection, parsed_args.dir)
    elif parsed_args.file:
        upload_file(collection, parsed_args.file)
    elif unknown_args:
        for arg in unknown_args:
            if os.path.isfile(arg):
                upload_file(collection, arg)
            elif os.path.isdir(arg):
                upload_directory(collection, arg)
            else:
                print(f"Skipping unknown argument: {arg}")
    else:
        print("Please provide a valid file or directory to upload")
        parser.print_help(sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()



# Jiggy Client
import os
from requests.auth import HTTPBasicAuth
from requests.packages.urllib3 import Retry
from requests.adapters import HTTPAdapter
import requests
from time import sleep
from .login import window_open

JIGGYBASE_HOST     = os.environ.get('JIGGYBASE_HOST', 'https://api.gpt-gateway.com')   # transition to jiggy.ai in progress

JB_KEY_FILE = os.path.expanduser('~') + '/.jiggybase'   # local file to store user entered apikey 

"""
Retry the following status codes:
500 Internal Server Error: The server encountered an error while processing the request.
502 Bad Gateway: The server, while acting as a gateway, received an invalid response from the upstream server.
503 Service Unavailable: The server is currently unable to handle the request due to maintenance or overload.
504 Gateway Timeout: The server, while acting as a gateway, did not receive a timely response from the upstream server.
"""
RETRY_STATUS_CODES = [500, 502, 503, 504]

class ClientError(Exception):
    """
    API returned 4xx client error
    """

class ServerError(Exception):
    """
    API returned 5xx Server error
    """

    
class JiggyBaseSession(requests.Session):
    def __init__(self, host=JIGGYBASE_HOST, api='gpt-gateway-v1',  *args, **kwargs):
        """
        Extend requests.Session with common GPTG authentication, retry, and exceptions.

        host: The url host prefix of the form "https:/api.gpt-gateway.com"
              if host arg is not set, will use 
                    JIGGYBASE_HOST environment variable or "api.gpt-gateway.com" as final default.

        api:  The api & version to use. defaults to 'gpt-gateway-v1'
                
        final url prefix are of the form "https://{host}/{api}"
        """
        super(JiggyBaseSession, self).__init__(*args, **kwargs)
        if not host.startswith('http'):
            host = f"https://{host}" if not host.startswith('localhost') else f'http://{host}'
        self.host = host
        if api:
            self.prefix_url = f"{host}/{api}"
        else:
            self.prefix_url = host            
        test_url = f"{self.prefix_url}/docs"
        if requests.head(test_url).status_code != 200:
            raise Exception(f"Invalid or unreachable api: {self.prefix_url}")
            
        self.bearer_token = None
        super(JiggyBaseSession, self).mount('https://',
                                            HTTPAdapter(max_retries=Retry(connect=5,
                                                                          read=5,
                                                                          status=5,
                                                                          redirect=2,
                                                                          backoff_factor=.001,
                                                                          status_forcelist=None)))

    def _set_bearer(self, jwt):
        self.bearer_token = jwt
        self.headers['Authorization'] = f"Bearer {jwt}"

    def _getjwt(self, key):        
        resp = requests.post(f"{JIGGYBASE_HOST}/gpt-gateway-v1/auth", json={'key': key})
        if resp.status_code == 200:
            self._set_bearer(resp.json()['jwt'])
        elif resp.status_code == 401:
            raise ClientError("Invalid API Key")
        else:
            raise ServerError(resp.content)

    def _auth(self):
        if 'JIGGYBASE_API_KEY' in os.environ:
            self._getjwt(os.environ['JIGGYBASE_API_KEY'])
        elif os.path.exists(JB_KEY_FILE):
            self._getjwt(open(JB_KEY_FILE).read())
        else:
            while True:
                window_open("https://jiggy.ai/authorize")
                key_input = input("Enter your JiggyBase API Key: ")
                if key_input[:4] == "jgy-":
                    # try using the key to see if it is valid
                    try:
                        self._getjwt(key_input)
                        # key validated, save to key file
                        open(JB_KEY_FILE, 'w').write(key_input)
                        os.chmod(JB_KEY_FILE, 0o600)  # -rw-------
                        break
                    except:
                        pass
                print("Invalid API Key")

        
    def request(self, method, url, *args, **kwargs):
        if not self.bearer_token:
            self._auth()
        url = self.prefix_url + url
        # support 'model' (pydantic BaseModel) arg which we convert to json parameter
        if 'model' in kwargs:
            kwargs['data'] = kwargs.pop('model').json()

        max_attempts = 7
        backoff_factor = 0.5

        for attempt in range(max_attempts):        
            try:        
                resp =  super(JiggyBaseSession, self).request(method, url, *args, **kwargs)
            except Exception as e:
                print(f"Exception: {e}")
                if attempt == max_attempts - 1:
                    raise
                continue

            if resp.status_code == 401:
                self.bearer_token = None
                del self.headers['Authorization']
                self._auth()
                if attempt == max_attempts - 1:
                    print("Unable to authenticate")
                    raise ValueError("Unable to authenticate")
                continue
            if resp.status_code not in RETRY_STATUS_CODES:
                break
            sleep_duration = backoff_factor * (2 ** attempt)
            sleep(sleep_duration)
            
        if resp.status_code >= 500:
            raise ServerError(resp.content)
        if resp.status_code >= 400:
            raise ClientError(resp.content)
        return resp





from typing import Optional, List
from pydantic import BaseConfig, EmailStr
import enum

from .collection import Collection, CollectionPostRequest

from .models.org import Org as OrgModel
from .models.org import OrgRole, OrgMember, OrgPatchRequest
from .models import PromptTask, PromptMessage, PromptTaskPostRequest, PromptTaskType

    
###
## Org
###
class Org(OrgModel):

    class Config(BaseConfig):
        extra = "allow"

    def __init__(self, session, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.session = session
        
    def create_collection(self, 
                          name: str, 
                          display_name: str,
                          plugin_auth: Optional[str] = "bearer",
                          description: Optional[str] = None) -> Collection:
        rsp = self.session.post(f"/orgs/{self.id}/collections", model=CollectionPostRequest(**locals()))
        return Collection(self.session, **rsp.json())

    def collections(self) -> list[Collection]:
        rsp = self.session.get(f"/orgs/{self.id}/collections")
        return [Collection(self.session, **c) for c in rsp.json()]

    def collection(self, name: str) -> Collection:
        collections = [c for c in self.collections() if c.name == name or c.display_name.lower() == name.lower()]
        if not collections:
            raise ValueError(f"Collection {name} not found")
        return collections[0]
    
    def members(self) -> List[OrgMember]:
        return [OrgMember(**tm) for tm in self.session.get(f'/orgs/{self.id}/members').json()]
    
    def add_member(self, email : EmailStr, role : OrgRole) -> OrgMember:
        """
        Invite the specified email to this org with the corresponding OrgRole.
        If the email does not have an account, the system will send an invite to this email.
        """
        rsp = self.session.post(f"/orgs/{self.id}/members", json={"email": email, 'role':role})
        return OrgMember(**rsp.json())

    def delete_member(self, email : str):
        """
        attempt to remove the specified name from this org
        """
        member = [m for m in self.members() if m.email == email]
        if not member:
            raise Exception(f"{email} not found in Org")
        self.session.delete(f"/orgs/{self.id}/members/{member[0].id}")

    def update(self, name: Optional[str] = None, description: Optional[str] = None) -> 'Org':
        """
        update the org name or description
        """
        patch_request = OrgPatchRequest(name=name, description=description)
        rsp = self.session.patch(f"/orgs/{self.id}", model=patch_request)
        updated_org_data = rsp.json()
        self.name = updated_org_data.get("name", self.name)
        self.description = updated_org_data.get("description", self.description)
        return self
        
    def prompt_tasks(self, name=None, version=None) -> list[PromptTask]:
        query = ''
        if name:
            query += f"?name={name}"
            if version:
                query += f"&version={version}"
        return [PromptTask(**pt) for pt in self.session.get(f"/orgs/{self.id}/prompt_tasks{query}").json()]

    def create_prompt_task(self, 
                           name       : str, 
                           version    : int, 
                           prompts    : List[PromptMessage],
                           type       : Optional[PromptTaskType] = None, 
                           description: Optional[str] = None, 
                           ) -> PromptTask:
        rsp = self.session.post(f"/orgs/{self.id}/prompt_tasks", model=PromptTaskPostRequest(**locals()))
        return PromptTask(**rsp.json())

    def update_prompt_task(self,
                           name       : str,
                           prompts    : List[PromptMessage]) -> PromptTask:
        pt = self.prompt_tasks(name=name)
        if not pt:
            raise ValueError(f"PromptTask {name} not found")
        pt = pt[0]
        self.create_prompt_task(name        = name, 
                                version     = pt.version+1, 
                                type        = pt.type, 
                                description = pt.description, 
                                prompts     = prompts)
        
    def get_prompt_task(self, prompt_task_id: int) -> PromptTask:
        
        return PromptTask(**self.session.get(f"/orgs/{self.id}/prompt_tasks/{prompt_task_id}").json())

    def delete_prompt_task(self, prompt_task_id: int):
        self.session.delete(f"/orgs/{self.id}/prompt_tasks/{prompt_task_id}")


"""
Config for chat-related functionality
"""

from pydantic import BaseModel, Field
from .chatmodel import ChatModelName
from typing import Optional, List


class CollectionChatConfig(BaseModel):
    """Config for Collection chat-related functionality"""    
    model_name          : str           = Field(description="The model associated with this configuration")
    rate_limit_messages : int           = Field(description="The total number of messages allowed in the rate limit time period across all collection users")
    rate_limit_hours    : int           = Field(description="The number of hours in the rate limit time period")
    prompt_task_id      : Optional[int] = Field(description="The prompt task id of the prompt to use for chat with this collection")


class PatchCollectionChatConfig(BaseModel):
    """Config for Collection chat-related functionality"""    
    prompt_task_id      : Optional[int] = Field(description="The prompt task id of the prompt to use for chat with this collection")




## Chat Completion Request objects sent to the chat completion endpoint

class Message(BaseModel):
    role: str
    content: str
    
class CompletionRequest(BaseModel):
    model: str
    messages: list[Message]
    max_tokens: Optional[int] = None
    temperature: float
    stream: bool




class Choice(BaseModel):
    finish_reason: str
    index: int
    message: Message

class Usage(BaseModel):
    completion_tokens: int
    prompt_tokens: int
    total_tokens: int

class ChatCompletion(BaseModel):
    id: str
    choices: List[Choice]
    created: int
    model: str
    object: str
    usage: Usage

    def __str__(self):
        return self.choices[0].message.content
    


class ChatUsage(BaseModel):
    """
    JiggyBase chat completion that was temporarily used before transitioning to the OpenAI ChatCompletion
    """
    collection_id:   int           = Field(description="The collection ID used to inform the completion")
    model:           str           = Field(description="The name of the model used to generate the completion")
    user_message:    str           = Field(description="The user's input message")
    input_tokens:    int           = Field(description="The number of input tokens sent to the model")
    output_tokens:   int           = Field(description="The number of output tokens generated by the model")
    messages_json:   Optional[str] = Field(description="JSON representation of the input messages sent to model endpoint")    
    completion:      Optional[str] = Field(description="The completion text returned by the model")
    created_at:      float         = Field(description="The epoch timestamp associated with the completion.")
    
    def __str__(self):
        return self.completion


import jiggybase
import sys


collection = jiggybase.JiggyBase().collection('hackernews-summary')

messages = [{'role':'user',  'content': 'articles about python 3.11'}]

for outstr in collection._chat_completion_stream_str(messages):
    sys.stdout.write(outstr)
    sys.stdout.flush()
print()



from typing import Optional, List
from pydantic import BaseModel, Field, HttpUrl, validator
from enum import Enum

from .chatmodel import ChatModelName

class ExtractMetadataConfig(BaseModel):
    """
    configuration for metadata extraction
    """
    created_at:  bool          = Field(True, description="Attempt to extract the created_at metadata field")
    author:      bool          = Field(True, description="Attempt to extract the author metadata field")
    title:       bool          = Field(True, description="Attempt to extract the title metadata field")
    model:       ChatModelName = Field(ChatModelName.gpt4, description="The name of the model to use for metadata extraction")



from typing import Optional
from pydantic import BaseModel, Field, EmailStr

###
## User
###

class User(BaseModel):
    id:              int = Field(description='Internal user_id')
    name:            str = Field(min_length=1, max_length=39, description='Unique name for the user.')
    email:           EmailStr  = Field(description='Email address for the user.')    
    auth0_userid:    str = Field(description='Auth0 userid.  This can be None for anonymous accounts created via api key')


class UserPostRequest(BaseModel):
    name:            str           = Field(min_length=1, max_length=39, description='Name for the user.')
    email:           EmailStr      = Field(description='Email address for the user.')
    

class UserPostPatchRequest(BaseModel):
    name:      Optional[str]      = Field(min_length=1, max_length=39, description='Unique name for the user.')



###
##  API Key Authentication
###

class AuthRequest(BaseModel):
    key : str = Field(description = "The API key")
    
class Jwt(BaseModel):
    jwt: str = Field(description='The JWT to used as bearer token')


###
## API Key
###


class  ApiKey(BaseModel):
    key:         str           = Field(default=None, description='The api key.')
    description: Optional[str] = Field(default=None, description='Optional user supplied description of the key.')
    created_at:  float         = Field(description='The epoch timestamp when the key was created.')
    last_used :  float         = Field(description='The epoch timestamp when the key was last used to create a JWT.')


class  ApiKeyRequest(BaseModel):
    description: Optional[str] = Field(default=None, description='Optional user supplied description of the key.')

class AllApiKeyResponse(BaseModel):
    items: list[ApiKey] = Field(description="List of all Api Keys")
        


from pydantic import BaseModel, Field
from typing import Optional


class ChunkConfig(BaseModel):
    """
    configuration for chunking policy
    """
    chunk_size:                 int = Field(200,   description="The target size of each text chunk in tokens")
    min_chunk_size_chars:       int = Field(350,   description="The minimum size of each text chunk in characters")
    min_chunk_length_to_embed:  int = Field(  5,   description="Discard chunks shorter than this")
    embeddings_batch_size:      int = Field(128,   description="The number of embeddings to request at a time")
    max_num_chunks:             int = Field(10000, description="The maximum number of chunks to generate from a text")



class PatchChunkConfigRequest(BaseModel):
    chunk_size:                Optional[int] = Field(description="The target size of each text chunk in tokens")
    min_chunk_size_chars:      Optional[int] = Field(description="The minimum size of each text chunk in characters")
    min_chunk_length_to_embed: Optional[int] = Field(description="Discard chunks shorter than this")
    embeddings_batch_size:     Optional[int] = Field(description="The number of embeddings to request at a time")
    max_num_chunks:            Optional[int] = Field(description="The maximum number of chunks to generate from a text")




from loguru import logger
import json

def extract_content_from_sse_bytes(sse_bytes: bytes) -> str:
    """
    Extracts the "content" value from the raw SSE byte string. This is the incremental model output text.
    """
    sse_str = sse_bytes.decode('utf-8')
    lines = sse_str.strip().split('\n\n')
    results = ""
    for line in lines:        
        if line.startswith("data:"):
            event_str = line.split(":", 1)[1].strip()
            if '[DONE]' in event_str:
                continue
            try:
                event_data = json.loads(event_str)
                content = event_data.get("choices", [])[0].get("delta", {}).get("content", "")
                results += content
            except (json.JSONDecodeError, IndexError, KeyError, TypeError) as e:
                logger.error(f"Failed to parse SSE event data for line: '{line}'")
                logger.error(f"Failed to parse SSE event data with err: {e}")
                continue
    return results



import webbrowser

def is_notebook() -> bool:
    """
    Returns True if code is executed in a notebook (Jupyter, Colab, QTconsole), False otherwise.
    https://stackoverflow.com/a/39662359
    """
    try:
        shell_class = get_ipython().__class__
        for parent_class in shell_class.__mro__:
            if parent_class.__name__ == "ZMQInteractiveShell":
                return True  # Jupyter notebook, Google colab or qtconsole
        return False
    except NameError:
        return False  # Probably standard Python interpreter

    
def window_open(url):    
    print(f"You can find your API Key here: {url}")
    if not is_notebook():
        print("(Attempting to open browser window...)")            
        webbrowser.open(url)
    else:
        pass
        # The following code seems to work to open a web page from a notebook,
        # but when successful it takes the user away to this page somewhat unexpectedly
        # so we will instead rely on the user clicking the above url so they
        # understand what is happening.
        #from IPython.display import Javascript        
        #display(Javascript('window.open("{url}");'.format(url=url)))



from pydantic import BaseModel, Field
from typing import Optional


class PineconeConfig(BaseModel):
    api_key: str = Field(..., description="Your Pinecone API Key")
    environment: str = Field(..., description="Your Pinecone environment")
    index: str = Field(..., description="Your Pinecone index")


class WeaviateConfig(BaseModel):
    host: str = Field(..., description="Your Weaviate host")
    port: int = Field(..., description="Your Weaviate port")
    index: str = Field(..., description="Your Weaviate index")
    username: Optional[str] = Field(None, description="Your Weaviate username")
    password: Optional[str] = Field(None, description="Your Weaviate password")
    scopes: Optional[str] = Field(None, description="Your Weaviate scopes")
    batch_size: Optional[int] = Field(None, description="Your Weaviate batch size")
    batch_dynamic: Optional[bool] = Field(None, description="Your Weaviate batch dynamic")
    batch_timeout_retries: Optional[int] = Field(None, description="Your Weaviate batch timeout retries")
    batch_num_workers: Optional[int] = Field(None, description="Your Weaviate batch number of workers")


class ZillizConfig(BaseModel):
    collection: str = Field(..., description="Your Zilliz collection")
    uri: str = Field(..., description="Your Zilliz URI")
    user: str = Field(..., description="Your Zilliz username")
    password: str = Field(..., description="Your Zilliz password")


class MilvusConfig(BaseModel):
    collection: str = Field(..., description="Your Milvus collection")
    host: str = Field(..., description="Your Milvus host")
    port: str = Field(default="19530", description="Your Milvus port")
    user: str = Field(None, description="Your Milvus username")
    password: str = Field(None, description="Your Milvus password")
    index_params: Optional[str] = Field(..., description="Custom index options for the collection.")
    search_params: Optional[str] = Field(..., description="Custom search options for the collection")
    consistency_level: Optional[str] = Field(..., description="Custom consistency level for the collection")
    
class QdrantConfig(BaseModel):
    url: str = Field(..., description="Your Qdrant URL")
    port: str = Field("6333", description="Your Qdrant port")
    grpc_port: str = Field("6334", description="Your Qdrant gRPC port")
    api_key: str = Field(..., description="Your Qdrant API key")
    collection: str = Field(..., description="Your Qdrant collection")


class RedisConfig(BaseModel):
    host: str = Field(..., description="Your Redis host")
    port: int = Field(..., description="Your Redis port")
    password: str = Field(None, description="Your Redis password")
    index_name: str = Field(..., description="Your Redis index name")
    doc_prefix: str = Field(..., description="Your Redis document prefix")
    distance_metric: Optional[str] = Field(None, description="Your Redis distance metric")
    index_type: Optional[str] = Field(None, description="Your Redis index type")





from pydantic import BaseModel, Field, HttpUrl
from typing import List



class PluginAuthConfigOAuth(BaseModel):
    """
    The Plugin Oauth configuration when plugin_auth == PluginAuthType.oauth
    """    
    client_url:                 HttpUrl     = Field(description="ChatGPT will direct the user’s browser to this url to log in to the plugin")
    authorization_url:          HttpUrl     = Field(description="After successful login ChatGPT will complete the OAuth flow by making a POST request to this URL")
    scope:                      str         = Field(description="The scope used for the OAuth flow")
    client_id:                  str         = Field(unique=True, index=True, description="The client id to send to OpenAI for the plugin")
    client_secret:              str         = Field(description="The client secret to send to OpenAI for the plugin")
    openai_verification_token:  str         = Field(description="The verification token specified by OpenAI to configure in the plugin")




class PluginBearerTokenConfig(BaseModel):
    """
    A list of bearer tokens that are authorized to access the /plugin api when plugin_auth == PluginAuthType.bearer
    """
    authorized_tokens : List[str]  =  []    


from typing import Optional, List
from pydantic import BaseModel, Field, HttpUrl, validator
from enum import Enum
import re
from time import time


class PluginAuthType(str, Enum):
    bearer :str = "bearer"
    none   :str = "none"
    oauth  :str = "oauth"
    

        
class CollectionPostRequest(BaseModel):
    """
    Used to create a new Collection service
    """
    name:         str                      = Field(description="The globally unique hostname for the collection. Subject to DNS naming rules. ")
    display_name: str                      = Field(description="The human friendly display name for the collection.")
    description:  Optional[str]            = Field(description="A description of the collection.")
    plugin_auth:  Optional[PluginAuthType] = Field(default=PluginAuthType.bearer, description="The authentication type to use for this collection's ChatGPT plugin.")


class CollectionPatchRequest(BaseModel):
    """
    Used to modify an existing service
    """
    description:  Optional[str] = Field(description="A description of the collection.")
    plugin_auth:  Optional[str] = Field(description='The auth plugin to use for this org.')
    display_name: Optional[str] = Field(description="The human friendly display name for the collection.")



class PluginAuthConfigOAuth(BaseModel):
    """
    The Plugin Oauth configuration as managed by GPT-Gateway
    """    
    client_url:                 HttpUrl     = Field(description="ChatGPT will direct the user’s browser to this url to log in to the plugin")
    authorization_url:          HttpUrl     = Field(description="After successful login ChatGPT will complete the OAuth flow by making a POST request to this URL")
    scope:                      str         = Field(description="The scope used for the OAuth flow")
    client_id:                  str         = Field(unique=True, index=True, description="The client id to send to OpenAI for the plugin")
    client_secret:              str         = Field(description="The client secret to send to OpenAI for the plugin")
    openai_verification_token:  str         = Field(description="The verification token specified by OpenAI to configure in the plugin")

class PatchPluginOAuthConfigRequest(BaseModel):
    openai_verification_token: str  




# Use DNS naming rules for convenience as they prohibit special characters, underscores, spaces, etc.
ALLOWED_NAME = re.compile("(?!-)[A-Z\d-]{1,63}(?<!-)$", re.IGNORECASE)   # REGEX for valid Name, based on DNS name component requirements
        
class Collection(BaseModel):
    """
    A managed collection of searchable documents exposed via as a ChatGPT plugin and a REST API.
    """
    id:          int            = Field(description="The unique ID of the service")
    name:        str            = Field(description="The DNS hostname for the service. The subdomain, not the FQDN.  Subject to DNS naming rules.")
    display_name: str           = Field(description="The human friendly display name for the collection.")    
    description: Optional[str]  = Field(description="A description of the collection.")
    fqdn:        str            = Field(description="The FQDN for the service, derived from the hostname and the domain name.")
    org_id:      int            = Field(description='The Org that owns this Service.')
    created_by:  int            = Field(description='The user_id that created this item.')
    updated_by:  int            = Field(description='The user_id that last modified this item.')
    plugin_auth: PluginAuthType = Field(description='The auth plugin to use for this org.')
    created_at:  float          = Field(description='The epoch timestamp when the collection was created.')
    updated_at:  float          = Field(description='The epoch timestamp when the collection was created.')

    @validator('name')
    def _name(cls, v):
        if len(v) > 63 or not v:
            raise ValueError(f'{v} is an invalid name. It must not be empty and is limited to 63 characters.')
        elif not bool(ALLOWED_NAME.match(v)):
            raise ValueError(f'{v} is an invalid name. It can only contain letters, numbers, and hyphens, and must not start or end with a hyphen.')
        return v



from typing import Optional, List
from pydantic import BaseModel, Field, HttpUrl, validator
from enum import Enum
from time import time


class PromptMessageRole(str, Enum):
    system              : str = "system"
    user                : str = "user"
    assistant           : str = "assistant"
    collection_context  : str = "collection_context"   # this is a special role that pulls in collection context in subsequent system messages
        
    def __str__(self):
        return str(self.value)

           
class PromptMessage(BaseModel):
    role      : PromptMessageRole  = Field(description="The role of the prompt")
    content   : str                = Field(description="The text of the prompt")
    position  : int                = Field(description="The position offset of the item in the prompt message list. 0-based, with negative values counting from the end.")


class PromptTaskType(str, Enum):
    chat            = "chat"             # iterative chat (e.g. chat with a model with user/assistant message history)
    collection_chat = "collection_chat"  # iterative chat augmented with retreival from a collection
    plugin_chat     = "plugin_chat"      # iterative chat augmented with tool use via plugin interface
    read_write      = "read_write"       # non-iterative task that involve reading a document and producing text or data

       
class PromptTask(BaseModel):
    id          :   int                      = Field(description="The unique ID of the prompt task")
    org_id      :   int                      = Field(description='The Org that owns this.')    
    name        :   str                      = Field(description="The name of the task. Unique within the org.")    
    type        :   Optional[PromptTaskType] = Field(description="The type of task")
    version     :   int                      = Field(description="The version of the task. Unique for a given org and task name.")
    description :   Optional[str]            = Field(description="The description of the task")
    prompts     :   List[PromptMessage]      = Field(description="The prompt messages used in the prompt task")
    created_at  :   float                    = Field(description='The epoch timestamp when the item was created.')

 
class PromptTaskPostRequest(BaseModel):
    name:          str                      = Field(description="The name of the task. Unique within the org.")
    version:       int                      = Field(description="The version of the task. Unique for a given org and task name.")
    type:          Optional[PromptTaskType] = Field(description="The type of task")        
    description:   Optional[str]            = Field(description="The description of the task")
    prompts:       List[PromptMessage]      = Field(description="The items that make up the prompt task")







from typing import Optional, Union, List
from pydantic import BaseModel, Field
from enum import Enum


class Source(str, Enum):
    email = "email"
    file = "file"
    chat = "chat"
    web  = "web"

class DocumentMetadata(BaseModel):
    source: Optional[Source] = None
    source_id: Optional[str] = None
    url: Optional[str] = None
    created_at: Optional[str] = None
    author: Union[str, list[str]] = None
    title: Optional[str] = None
    description: Optional[str] = None
    

class DocumentChunkMetadata(DocumentMetadata):
    document_id: Optional[str] = None


class DocumentChunk(BaseModel):
    id: Optional[str] = None
    text: str
    metadata: DocumentChunkMetadata
    embedding: Optional[list[float]] = None
    token_count: Optional[int] = None

    def __str__(self):
        if len(self.text) > 100:
            text = self.text[:100] + '...'
        else: 
            text = self.text
        text = text.replace('\n', ' ')
        estr = "DocumentChunk("
        if self.id is not None:
            estr += f"id={self.id}, "
        if self.metadata is not None:
            estr += f"metadata={str(self.metadata)}, "
        if self.embedding is not None:
            estr += f"embedding=dim{len(self.embedding)}, "
        estr += f"text={text})"
        return estr

    
class DocumentChunkWithScore(DocumentChunk):
    score: float


class Document(BaseModel):
    id: Optional[str]
    text: str
    metadata: Optional[DocumentMetadata] = None
    mimetype: Optional[str] = None
    token_count: Optional[int] = None


class DocumentWithChunks(Document):
    chunks: List[DocumentChunk]

class DocumentMetadataFilter(BaseModel):
    document_id: Optional[str] = None
    source: Optional[Source] = None
    source_id: Optional[str] = None
    author: Optional[str] = None
    start_date: Optional[str] = None  # any date string format
    end_date: Optional[str] = None  # any date string format
    title: Optional[str] = None


class Query(BaseModel):
    query: str
    filter: Optional[DocumentMetadataFilter] = None
    top_k: Optional[int] = 3

class QueryWithEmbedding(Query):
    embedding: list[float]


class QueryResult(BaseModel):
    query: str
    results: list[DocumentChunkWithScore]


class UpsertRequest(BaseModel):
    documents: List[Document]


class UpsertResponse(BaseModel):
    ids: List[str]


class QueryRequest(BaseModel):
    queries: List[Query]


class QueryResponse(BaseModel):
    results: List[QueryResult]


class DeleteRequest(BaseModel):
    ids: Optional[List[str]] = None
    filter: Optional[DocumentMetadataFilter] = None
    delete_all: Optional[bool] = False


class DeleteResponse(BaseModel):
    success: bool


class Accounting(BaseModel):
    chunk_count: int
    doc_count: int
    page_count: int



class DocChunksResponse(BaseModel):
    docs       : list[list[DocumentChunk]] = Field(..., description="A list of documents, each containing a list of chunks")
    next_index : int                       = Field(..., description="The index of the next document to return.  Use this value as the index parameter in the next request to get the next set of documents")  
          